@@ -30,6 +30,10 @@
 #include "e-data-server-ui-marshal.h"
 #include "e-source-selector.h"
 
+#define E_SOURCE_SELECTOR_GET_PRIVATE(obj) \
+	(G_TYPE_INSTANCE_GET_PRIVATE \
+	((obj), E_TYPE_SOURCE_SELECTOR, ESourceSelectorPrivate))
+
 struct _ESourceSelectorPrivate {
 	ESourceList *list;
 
@@ -55,18 +59,79 @@ typedef struct {
 	gboolean selection_changed;
 } ESourceSelectorRebuildData;
 
+enum {
+	PROP_0,
+	PROP_SOURCE_LIST
+};
+
 enum {
 	SELECTION_CHANGED,
 	PRIMARY_SELECTION_CHANGED,
 	POPUP_EVENT,
+	DATA_DROPPED,
 	NUM_SIGNALS
 };
 static unsigned int signals[NUM_SIGNALS] = { 0 };
 
-static gboolean selector_popup_menu (GtkWidget *widget);
-
 G_DEFINE_TYPE (ESourceSelector, e_source_selector, GTK_TYPE_TREE_VIEW)
 
+/* safe-toggle-renderer definition - it'll not call 'toggled' signal on 'activate', when mouse is not over the toggle */
+
+typedef struct _ECellRendererSafeToggle {
+	GtkCellRendererToggle parent;
+} ECellRendererSafeToggle;
+
+typedef struct _ECellRendererSafeToggleClass {
+	GtkCellRendererToggleClass parent_class;
+} ECellRendererSafeToggleClass;
+
+GType e_cell_renderer_safe_toggle_get_type (void);
+
+G_DEFINE_TYPE (ECellRendererSafeToggle, e_cell_renderer_safe_toggle, GTK_TYPE_CELL_RENDERER_TOGGLE)
+
+static gboolean
+safe_toggle_activate (GtkCellRenderer      *cell,
+		      GdkEvent             *event,
+		      GtkWidget            *widget,
+		      const gchar          *path,
+		      GdkRectangle         *background_area,
+		      GdkRectangle         *cell_area,
+		      GtkCellRendererState  flags)
+{
+	if (event->type == GDK_BUTTON_PRESS && cell_area) {
+		GdkRegion *reg = gdk_region_rectangle (cell_area);
+
+		if (!gdk_region_point_in (reg, event->button.x, event->button.y)) {
+			gdk_region_destroy (reg);
+			return FALSE;
+		}
+
+		gdk_region_destroy (reg);
+	}
+
+	return GTK_CELL_RENDERER_CLASS (e_cell_renderer_safe_toggle_parent_class)->activate (cell, event, widget, path, background_area, cell_area, flags);
+}
+
+static void
+e_cell_renderer_safe_toggle_class_init (ECellRendererSafeToggleClass *klass)
+{
+	GtkCellRendererClass *rndr_class;
+	
+	rndr_class = GTK_CELL_RENDERER_CLASS (klass);
+	rndr_class->activate = safe_toggle_activate;
+}
+
+static void
+e_cell_renderer_safe_toggle_init (ECellRendererSafeToggle *obj)
+{
+}
+
+static GtkCellRenderer *
+e_cell_renderer_safe_toggle_new (void)
+{
+	return g_object_new (e_cell_renderer_safe_toggle_get_type (), NULL);
+}
+
 /* Selection management.  */
 
 static GHashTable *
@@ -388,15 +453,6 @@ list_changed_callback (ESourceList *list,
 							  selector);
 }
 
-static void
-setup_model (ESourceSelector *selector)
-{
-	rebuild_model (selector);
-
-	g_signal_connect_object (selector->priv->list, "changed", G_CALLBACK (list_changed_callback), G_OBJECT (selector), 0);
-}
-
-
 /* Data functions for rendering the model.  */
 
 static void
@@ -547,6 +603,31 @@ selection_func (GtkTreeSelection *selection,
 
 /* Callbacks.  */
 
+static void
+text_cell_edited_cb (ESourceSelector *selector,
+                     const gchar *path_string,
+                     const gchar *new_name)
+{
+	GtkTreeView *tree_view;
+	GtkTreeModel *model;
+	GtkTreePath *path;
+	GtkTreeIter iter;
+	ESource *source;
+
+	tree_view = GTK_TREE_VIEW (selector);
+	model = gtk_tree_view_get_model (tree_view);
+	path = gtk_tree_path_new_from_string (path_string);
+
+	gtk_tree_model_get_iter (model, &iter, path);
+	gtk_tree_model_get (model, &iter, 0, &source, -1);
+	gtk_tree_path_free (path);
+
+	g_return_if_fail (E_IS_SOURCE (source));
+
+	if (new_name != NULL && *new_name != '\0')
+		e_source_set_name (source, new_name);
+}
+
 static void
 cell_toggled_callback (GtkCellRendererToggle *renderer,
 		       const char *path_string,
@@ -647,18 +728,6 @@ row_expanded_callback (GtkTreeView *treeview, GtkTreeIter *iter, GtkTreePath *pa
 	return FALSE;
 }
 
-static gboolean
-selector_popup_menu (GtkWidget *widget)
-{
-	ESourceSelector *selector = E_SOURCE_SELECTOR (widget);
-	ESource *source;
-	gboolean res = FALSE;
-
-	source = e_source_selector_peek_primary_selection (selector);
-	g_signal_emit (selector, signals[POPUP_EVENT], 0, source, NULL, &res);
-	return res;
-}
-
 static gboolean
 selector_button_press_event (GtkWidget *widget, GdkEventButton *event, ESourceSelector *selector)
 {
@@ -711,14 +780,64 @@ selector_button_press_event (GtkWidget *widget, GdkEventButton *event, ESourceSe
 /* GObject methods.  */
 
 static void
-e_source_selector_dispose (GObject *object)
+source_selector_set_source_list (ESourceSelector *selector,
+                                 ESourceList *source_list)
+{
+	g_return_if_fail (E_IS_SOURCE_LIST (source_list));
+	g_return_if_fail (selector->priv->list == NULL);
+
+	selector->priv->list = g_object_ref (source_list);
+
+	rebuild_model (selector);
+
+	g_signal_connect_object (
+		source_list, "changed",
+		G_CALLBACK (list_changed_callback),
+		G_OBJECT (selector), 0);
+
+	gtk_tree_view_expand_all (GTK_TREE_VIEW (selector));
+}
+
+static void
+source_selector_set_property (GObject *object,
+                              guint property_id,
+                              const GValue *value,
+                              GParamSpec *pspec)
+{
+	switch (property_id) {
+		case PROP_SOURCE_LIST:
+			source_selector_set_source_list (
+				E_SOURCE_SELECTOR (object),
+				g_value_get_object (value));
+			return;
+	}
+
+	G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
+}
+
+static void
+source_selector_get_property (GObject *object,
+                              guint property_id,
+                              GValue *value,
+                              GParamSpec *pspec)
+{
+	switch (property_id) {
+		case PROP_SOURCE_LIST:
+			g_value_set_object (
+				value, e_source_selector_get_source_list (
+				E_SOURCE_SELECTOR (object)));
+			return;
+	}
+
+	G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
+}
+
+static void
+source_selector_dispose (GObject *object)
 {
 	ESourceSelectorPrivate *priv = E_SOURCE_SELECTOR (object)->priv;
 
-	if (priv->selected_sources != NULL) {
-		g_hash_table_destroy (priv->selected_sources);
-		priv->selected_sources = NULL;
-	}
+	g_hash_table_remove_all (priv->selected_sources);
 
 	if (priv->rebuild_model_idle_id != 0) {
 		g_source_remove (priv->rebuild_model_idle_id);
@@ -737,19 +856,176 @@ e_source_selector_dispose (GObject *object)
 
 	clear_saved_primary_selection (E_SOURCE_SELECTOR (object));
 
-	(* G_OBJECT_CLASS (e_source_selector_parent_class)->dispose) (object);
+	/* Chain up to parent's dispose() method. */
+	G_OBJECT_CLASS (e_source_selector_parent_class)->dispose (object);
 }
 
 static void
-e_source_selector_finalize (GObject *object)
+source_selector_finalize (GObject *object)
 {
 	ESourceSelectorPrivate *priv = E_SOURCE_SELECTOR (object)->priv;
 
-	g_free (priv);
+	g_hash_table_destroy (priv->selected_sources);
 
-	(* G_OBJECT_CLASS (e_source_selector_parent_class)->finalize) (object);
+	/* Chain up to parent's finalize() method. */
+	G_OBJECT_CLASS (e_source_selector_parent_class)->finalize (object);
 }
 
+static void
+source_selector_drag_leave (GtkWidget *widget,
+                            GdkDragContext *context,
+                            guint time_)
+{
+	GtkTreeView *tree_view;
+	GtkTreeViewDropPosition pos;
+
+	tree_view = GTK_TREE_VIEW (widget);
+	pos = GTK_TREE_VIEW_DROP_BEFORE;
+
+	gtk_tree_view_set_drag_dest_row (tree_view, NULL, pos);
+}
+
+static gboolean
+source_selector_drag_motion (GtkWidget *widget,
+                             GdkDragContext *context,
+                             gint x,
+                             gint y,
+                             guint time_)
+{
+	GtkTreeView *tree_view;
+	GtkTreeModel *model;
+	GtkTreePath *path = NULL;
+	GtkTreeIter iter;
+	GtkTreeViewDropPosition pos;
+	GdkDragAction action = 0;
+	gpointer object = NULL;
+
+	tree_view = GTK_TREE_VIEW (widget);
+	model = gtk_tree_view_get_model (tree_view);
+
+	if (!gtk_tree_view_get_dest_row_at_pos (tree_view, x, y, &path, NULL))
+		goto exit;
+
+	if (!gtk_tree_model_get_iter (model, &iter, path))
+		goto exit;
+
+	gtk_tree_model_get (model, &iter, 0, &object, -1);
+
+	if (E_IS_SOURCE_GROUP (object) || e_source_get_readonly (object))
+		goto exit;
+
+	pos = GTK_TREE_VIEW_DROP_INTO_OR_BEFORE;
+	gtk_tree_view_set_drag_dest_row (tree_view, path, pos);
+
+	if (context->actions & GDK_ACTION_MOVE)
+		action = GDK_ACTION_MOVE;
+	else
+		action = context->suggested_action;
+
+exit:
+	if (path != NULL)
+		gtk_tree_path_free (path);
+
+	if (object != NULL)
+		g_object_unref (object);
+
+	gdk_drag_status (context, action, time_);
+
+	return TRUE;
+}
+
+static gboolean
+source_selector_drag_drop (GtkWidget *widget,
+                           GdkDragContext *context,
+                           gint x,
+                           gint y,
+                           guint time_)
+{
+	GtkTreeView *tree_view;
+	GtkTreeModel *model;
+	GtkTreePath *path;
+	GtkTreeIter iter;
+	gboolean drop_zone;
+	gboolean valid;
+	gpointer object;
+
+	tree_view = GTK_TREE_VIEW (widget);
+	model = gtk_tree_view_get_model (tree_view);
+
+	if (!gtk_tree_view_get_path_at_pos (
+		tree_view, x, y, &path, NULL, NULL, NULL))
+		return FALSE;
+
+	valid = gtk_tree_model_get_iter (model, &iter, path);
+	gtk_tree_path_free (path);
+	g_return_val_if_fail (valid, FALSE);
+
+	gtk_tree_model_get (model, &iter, 0, &object, -1);
+	drop_zone = E_IS_SOURCE (object);
+	g_object_unref (object);
+
+	return drop_zone;
+}
+
+static void
+source_selector_drag_data_received (GtkWidget *widget,
+                                    GdkDragContext *context,
+                                    gint x,
+                                    gint y,
+                                    GtkSelectionData *selection_data,
+                                    guint info,
+                                    guint time_)
+{
+	GtkTreeView *tree_view;
+	GtkTreeModel *model;
+	GtkTreePath *path = NULL;
+	GtkTreeIter iter;
+	gpointer object = NULL;
+	gboolean delete;
+	gboolean success = FALSE;
+
+	tree_view = GTK_TREE_VIEW (widget);
+	model = gtk_tree_view_get_model (tree_view);
+	delete = (context->action == GDK_ACTION_MOVE);
+
+	if (!gtk_tree_view_get_dest_row_at_pos (tree_view, x, y, &path, NULL))
+		goto exit;
+
+	if (!gtk_tree_model_get_iter (model, &iter, path))
+		goto exit;
+
+	gtk_tree_model_get (model, &iter, 0, &object, -1);
+
+	if (!E_IS_SOURCE (object) || e_source_get_readonly (object))
+		goto exit;
+
+	g_signal_emit (
+		widget, signals[DATA_DROPPED], 0, selection_data,
+		object, context->action, info, &success);
+
+exit:
+	if (path != NULL)
+		gtk_tree_path_free (path);
+
+	if (object != NULL)
+		g_object_unref (object);
+
+	gtk_drag_finish (context, success, delete, time_);
+}
+
+static gboolean
+source_selector_popup_menu (GtkWidget *widget)
+{
+	ESourceSelector *selector;
+	ESource *source;
+	gboolean res = FALSE;
+
+	selector = E_SOURCE_SELECTOR (widget);
+	source = e_source_selector_peek_primary_selection (selector);
+	g_signal_emit (selector, signals[POPUP_EVENT], 0, source, NULL, &res);
+
+	return res;
+}
 
 /* Initialization.  */
 static gboolean
@@ -765,13 +1041,34 @@ ess_bool_accumulator(GSignalInvocationHint *ihint, GValue *out, const GValue *in
 static void
 e_source_selector_class_init (ESourceSelectorClass *class)
 {
-	GObjectClass *object_class = G_OBJECT_CLASS (class);
-	GtkWidgetClass *widget_class = GTK_WIDGET_CLASS (class);
+	GObjectClass *object_class;
+	GtkWidgetClass *widget_class;
 
-	object_class->dispose  = e_source_selector_dispose;
-	object_class->finalize = e_source_selector_finalize;
-	widget_class->popup_menu = selector_popup_menu;
+	g_type_class_add_private (class, sizeof (ESourceSelectorPrivate));
 
+	object_class = G_OBJECT_CLASS (class);
+	object_class->set_property = source_selector_set_property;
+	object_class->get_property = source_selector_get_property;
+	object_class->dispose  = source_selector_dispose;
+	object_class->finalize = source_selector_finalize;
+
+	widget_class = GTK_WIDGET_CLASS (class);
+	widget_class->drag_leave = source_selector_drag_leave;
+	widget_class->drag_motion = source_selector_drag_motion;
+	widget_class->drag_drop = source_selector_drag_drop;
+	widget_class->drag_data_received = source_selector_drag_data_received;
+	widget_class->popup_menu = source_selector_popup_menu;
+
+	g_object_class_install_property (
+		object_class,
+		PROP_SOURCE_LIST,
+		g_param_spec_object (
+			"source-list",
+			NULL,
+			NULL,
+			E_TYPE_SOURCE_LIST,
+			G_PARAM_READWRITE |
+			G_PARAM_CONSTRUCT_ONLY));
 
 	signals[SELECTION_CHANGED] =
 		g_signal_new ("selection_changed",
@@ -799,6 +1096,18 @@ e_source_selector_class_init (ESourceSelectorClass *class)
 			      e_data_server_ui_marshal_BOOLEAN__OBJECT_BOXED,
 			      G_TYPE_BOOLEAN, 2, G_TYPE_OBJECT,
 			      GDK_TYPE_EVENT|G_SIGNAL_TYPE_STATIC_SCOPE);
+	signals[DATA_DROPPED] =
+		g_signal_new ("data_dropped",
+			      G_OBJECT_CLASS_TYPE (object_class),
+			      G_SIGNAL_RUN_LAST,
+			      G_STRUCT_OFFSET (ESourceSelectorClass, data_dropped),
+			      NULL, NULL,
+			      e_data_server_ui_marshal_BOOLEAN__BOXED_OBJECT_FLAGS_UINT,
+			      G_TYPE_BOOLEAN, 4,
+			      GTK_TYPE_SELECTION_DATA | G_SIGNAL_TYPE_STATIC_SCOPE,
+			      E_TYPE_SOURCE,
+			      GDK_TYPE_DRAG_ACTION,
+			      G_TYPE_UINT);
 }
 
 static gboolean
@@ -838,13 +1147,16 @@ e_source_selector_init (ESourceSelector *selector)
 	GtkTreeViewColumn *column;
 	GtkCellRenderer *cell_renderer;
 	GtkTreeSelection *selection;
+	GtkTreeView *tree_view;
 
-	priv = g_new0 (ESourceSelectorPrivate, 1);
-	selector->priv = priv;
+	selector->priv = E_SOURCE_SELECTOR_GET_PRIVATE (selector);
+	priv = selector->priv;
 
-	gtk_tree_view_set_search_column (GTK_TREE_VIEW (selector), 0);
-	gtk_tree_view_set_search_equal_func (GTK_TREE_VIEW (selector), group_search_function, NULL, NULL);
-	gtk_tree_view_set_enable_search (GTK_TREE_VIEW (selector), TRUE);
+	tree_view = GTK_TREE_VIEW (selector);
+
+	gtk_tree_view_set_search_column (tree_view, 0);
+	gtk_tree_view_set_search_equal_func (tree_view, group_search_function, NULL, NULL);
+	gtk_tree_view_set_enable_search (tree_view, TRUE);
 
 	g_signal_connect (G_OBJECT (selector), "button_press_event",
 			  G_CALLBACK (selector_button_press_event), selector);
@@ -856,31 +1168,33 @@ e_source_selector_init (ESourceSelector *selector)
 	priv->selected_sources = create_selected_sources_hash ();
 
 	priv->tree_store = gtk_tree_store_new (1, G_TYPE_OBJECT);
-	gtk_tree_view_set_model (GTK_TREE_VIEW (selector), GTK_TREE_MODEL (priv->tree_store));
+	gtk_tree_view_set_model (tree_view, GTK_TREE_MODEL (priv->tree_store));
 
 	column = gtk_tree_view_column_new ();
-	gtk_tree_view_append_column (GTK_TREE_VIEW (selector), column);
+	gtk_tree_view_append_column (tree_view, column);
 
 	cell_renderer = gtk_cell_renderer_pixbuf_new ();
 	g_object_set (G_OBJECT (cell_renderer), "mode", GTK_CELL_RENDERER_MODE_ACTIVATABLE, NULL);
 	gtk_tree_view_column_pack_start (column, cell_renderer, FALSE);
 	gtk_tree_view_column_set_cell_data_func (column, cell_renderer, (GtkTreeCellDataFunc) pixbuf_cell_data_func, selector, NULL);
-	cell_renderer = gtk_cell_renderer_toggle_new ();
+	cell_renderer = e_cell_renderer_safe_toggle_new ();
 	gtk_tree_view_column_pack_start (column, cell_renderer, FALSE);
 	gtk_tree_view_column_set_cell_data_func (column, cell_renderer, (GtkTreeCellDataFunc) toggle_cell_data_func, selector, NULL);
 	g_signal_connect (cell_renderer, "toggled", G_CALLBACK (cell_toggled_callback), selector);
 
 	cell_renderer = gtk_cell_renderer_text_new ();
-	g_object_set (G_OBJECT (cell_renderer), "mode", GTK_CELL_RENDERER_MODE_ACTIVATABLE, NULL);
 	g_object_set (G_OBJECT (cell_renderer), "ellipsize", PANGO_ELLIPSIZE_END, NULL);
+	g_signal_connect_swapped (
+		cell_renderer, "edited",
+		G_CALLBACK (text_cell_edited_cb), selector);
 	gtk_tree_view_column_pack_start (column, cell_renderer, TRUE);
 	gtk_tree_view_column_set_cell_data_func (column, cell_renderer, (GtkTreeCellDataFunc) text_cell_data_func, selector, NULL);
 
-	selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (selector));
+	selection = gtk_tree_view_get_selection (tree_view);
 	gtk_tree_selection_set_select_function (selection, (GtkTreeSelectionFunc) selection_func, selector, NULL);
 	g_signal_connect_object (selection, "changed", G_CALLBACK (selection_changed_callback), G_OBJECT (selector), 0);
 
-	gtk_tree_view_set_headers_visible (GTK_TREE_VIEW (selector), FALSE);
+	gtk_tree_view_set_headers_visible (tree_view, FALSE);
 
 	g_signal_connect (G_OBJECT (selector), "test-collapse-row", G_CALLBACK (test_collapse_row_callback), selector);
 	g_signal_connect (G_OBJECT (selector), "row-expanded", G_CALLBACK (row_expanded_callback), selector);
@@ -901,20 +1215,10 @@ e_source_selector_init (ESourceSelector *selector)
 GtkWidget *
 e_source_selector_new (ESourceList *list)
 {
-	ESourceSelector *selector;
-
 	g_return_val_if_fail (E_IS_SOURCE_LIST (list), NULL);
 
-	selector = g_object_new (e_source_selector_get_type (), NULL);
-
-	selector->priv->list = list;
-	g_object_ref (list);
-
-	setup_model (selector);
-
-	gtk_tree_view_expand_all (GTK_TREE_VIEW (selector));
-
-	return GTK_WIDGET (selector);
+	return g_object_new (
+		E_TYPE_SOURCE_SELECTOR, "source-list", list, NULL);
 }
 
 /**
@@ -928,7 +1232,7 @@ e_source_selector_new (ESourceList *list)
 ESourceList *
 e_source_selector_get_source_list (ESourceSelector *selector)
 {
-	g_return_val_if_fail (E_IS_SOURCE_LIST (selector), NULL);
+	g_return_val_if_fail (E_IS_SOURCE_SELECTOR (selector), NULL);
 
 	return selector->priv->list;
 }
@@ -1152,6 +1456,64 @@ e_source_selector_source_is_selected (ESourceSelector *selector,
 	return source && source_is_selected (selector, source);
 }
 
+/**
+ * e_source_selector_edit_primary_selection:
+ * @selector: An #ESourceSelector widget
+ *
+ * Allows the user to rename the primary selected source by opening an
+ * entry box directly in @selector.
+ **/
+void
+e_source_selector_edit_primary_selection (ESourceSelector *selector)
+{
+	GtkTreeRowReference *reference;
+	GtkTreeSelection *selection;
+	GtkTreeViewColumn *column;
+	GtkCellRenderer *renderer;
+	GtkTreeView *tree_view;
+	GtkTreeModel *model;
+	GtkTreePath *path = NULL;
+	GtkTreeIter iter;
+	GList *list;
+
+	g_return_if_fail (E_IS_SOURCE_SELECTOR (selector));
+
+	tree_view = GTK_TREE_VIEW (selector);
+	column = gtk_tree_view_get_column (tree_view, 0);
+	reference = selector->priv->saved_primary_selection;
+	selection = gtk_tree_view_get_selection (tree_view);
+
+	if (reference != NULL)
+		path = gtk_tree_row_reference_get_path (reference);
+	else if (gtk_tree_selection_get_selected (selection, &model, &iter))
+		path = gtk_tree_model_get_path (model, &iter);
+
+	if (path == NULL)
+		return;
+
+	/* XXX Because we stuff three renderers in a single column,
+	 *     we have to manually hunt for the text renderer. */
+	list = gtk_tree_view_column_get_cell_renderers (column);
+	while (list != NULL) {
+		renderer = list->data;
+		if (GTK_IS_CELL_RENDERER_TEXT (renderer))
+			break;
+		list = g_list_delete_link (list, list);
+	}
+	g_list_free (list);
+
+	/* Make the text cell renderer editable, but only temporarily.
+	 * We don't want editing to be activated by simply clicking on
+	 * the source name.  Too easy for accidental edits to occur. */
+	g_object_set (renderer, "editable", TRUE, NULL);
+	gtk_tree_view_expand_to_path (tree_view, path);
+	gtk_tree_view_set_cursor_on_cell (
+		tree_view, path, column, renderer, TRUE);
+	g_object_set (renderer, "editable", FALSE, NULL);
+
+	gtk_tree_path_free (path);
+}
+
 /**
  * e_source_selector_peek_primary_selection:
  * @selector: An #ESourceSelector widget
